`timescale 1ps / 1 ps

module halfadd (a, b, q, cout);
	
	input	a;		// in value 0
	input	b;		// in value 1
	output	q;		// out value
	output	cout;	// carry out

	assign q	= a^b;
	assign cout = (a&b);

endmodule

module fulladd (a, b, cin, q, cout);
	
	input	a;		// in value 0
	input	b;		// in value 1
	input	cin;	// carry in
	output	q;		// out value
	output	cout;	// carry out

	assign q	= a^b^cin;
	assign cout = (a&b) | (a&cin) | (b&cin);

endmodule

module carryadd (a, cin, q, cout);
	
	input	a;		// in value 0
	input	cin;	// carry in
	output	q;		// out value
	output	cout;	// carry out

	assign q	= a^cin;
	assign cout = (a&cin);

endmodule

module halfsub(a,b,q,bout);
	input	a;
	input	b;
	output	q;
	output	bout;
	
	assign q	= a^b;
	assign bout	= (~a & b);
endmodule

module fullsub (a, b, bin, q, bout);
	
	input	a;		// in value 0
	input	b;		// in value 1
	input	bin;	// borrow in
	output	q;		// out value
	output	bout;	// borrow out

	assign q	= a^b^bin;
	assign bout = (~a & b)|(~(a ^ b) & bin);

endmodule

// q = a - b
module sub (a, b, q, borrow);
	
	parameter Width	= 16;
	
	input	[Width-1:0]	a;
	input	[Width-1:0]	b;
	output 	[Width-1:0]	q;
	output			borrow;
	wire	[Width-2:0]	c;

	halfsub sub0 (a[0], b[0], q[0], c[0]);
	generate
		genvar i;
		for(i = 1; i < Width-1; i=i+1)
		begin: fullsub_gen
			fullsub add_ (a[i], b[i], c[i-1], q[i], c[i]);
		end
		begin: fullsub_gen_last
			fullsub sub_last (a[Width-1], b[Width-1], c[Width-2], q[Width-1], borrow);
		end
	endgenerate
	
endmodule

// q = a + 1
module inc(a, q);
	parameter Width	= 16;
	
	input	[Width-1:0]	a;
	output	[Width-1:0]	q;
	wire	[Width-2:0]	c;
	wire			na;
	
	
	halfadd add0 (a[0], 1'b1, q[0], c[0]);
	generate
		genvar i;
		for(i = 1; i < Width-1; i=i+1)
		begin: carryadd_gen
			carryadd add_ (a[i], c[i-1], q[i], c[i]);
		end
		begin: carryadd_gen_last
			carryadd add_last (a[Width-1], c[Width-2], q[Width-1], na);
		end
	endgenerate
endmodule

// q = -a
module minus(a, q);

	parameter Width	= 16;
	
	input	[Width-1:0]	a;
	output 	[Width-1:0]	q;
	wire	[Width-1:0]	t;
	inc #(.Width(Width))	inc0(t, q);
	assign t = ~a;
	
endmodule

// q = a + b
module add (a, b, q, carry);
	
	parameter Width	= 16;
	
	input	[Width-1:0]	a;
	input	[Width-1:0]	b;
	output 	[Width-1:0]	q;
	output			carry;
	wire	[Width-2:0]	c;

	halfadd add0 (a[0], b[0], q[0], c[0]);
	generate
		genvar i;
		for(i = 1; i < Width-1; i=i+1)
		begin: fulladd_gen
			fulladd add_ (a[i], b[i], c[i-1], q[i], c[i]);
		end
		begin: fulladd_gen_last
			fulladd add_last (a[Width-1], b[Width-1], c[Width-2], q[Width-1], carry);
		end
	endgenerate
endmodule

/*
module mul_booth(a, b, q);

	parameter Width	= 16;

	input	[2:0]	a;
	input	[Width-1:0]	b;
	output	[Width:0]	q;
	wire	[Width:0]	sl_out;
	wire	[Width-1:0]	sl_in;
	wire	[Width:0] minus_in;
	wire	[Width:0] minus_out;
	
	always begin
		case (a)
			3'b000, 3'b111:	q <= 0;
			3'b001, 3'b010:	q <= {1'b0, b[Width-1:0]};
			3'b011:			begin
							sl_in<= b;
							q <= sl_out;
							end
			3'b100:			begin
							minus_in <= {1'b0, b[Width-1:0]};
							q <= minus_out;
							end
			3'b101, 3'b110:	begin
							sl_in <= b;
							minus_in <= sl_out;
							q <= minus_out;
							end
		endcase
	end
	
	assign sl_out = {sl_in[Width-1:0],1'b0};
	minus #(.Width(Width+1)) minus0	(minus_in, minus_out);
endmodule
*/
/*
module mul (a, b, q);
	input	[3:0]	a;
	input	[3:0]	b;
	input	[7:0]	q;
	wire 	[3:0]	t [0:3];
	
	assign t[0] = a & b[0];
	assign t[1] = a & b[1];
	assign t[2] = a & b[2];
	assign t[3] = a & b[3];
	
	
endmodule
*/